import numpy as np

from sklearn.datasets import make_regression
from sklearn.tree import DecisionTreeRegressor
from sklearn.utils.estimator_checks import check_estimator

from mubofo import MultioutputDecisionTree

def test_is_estimator():
    check_estimator(MultioutputDecisionTree())

def test_predictions():
    X, Y = make_regression(n_samples=200, n_features=40, n_targets=5)
    base = DecisionTreeRegressor(max_depth=3, random_state=123).fit(X, Y)
    model = MultioutputDecisionTree(max_depth=3, random_state=123).fit(X, Y)

    assert np.allclose(base.predict(X), model.predict(X))

def test_1d_importances():
    X, Y = make_regression(n_samples=200, n_features=40, n_targets=1)
    base = DecisionTreeRegressor(max_depth=3, random_state=123).fit(X, Y)
    model = MultioutputDecisionTree(max_depth=3, random_state=123).fit(X, Y)

    assert np.allclose(base.feature_importances_, model.feature_importances_)



import numpy as np

from sklearn.datasets import make_regression
from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import r2_score
from sklearn.utils.estimator_checks import check_estimator

from mubofo import MultioutputBoostedForest, MultioutputRandomForest

def test_MultioutputBoostedForest_is_estimator():
    check_estimator(MultioutputBoostedForest())

def test_MultioutputBoostedForest_learns_well():
    X, Y = make_regression(n_samples=10000, n_features=40, n_targets=20)
    Y = (Y - Y.mean(axis=0)) / Y.std(axis=0)

    model = MultioutputBoostedForest(max_depth=2).fit(X, Y)
    score = r2_score(Y, model.predict(X))

    assert score > 0.9

def test_MultioutputRandomForest_is_estimator():
    check_estimator(MultioutputRandomForest())

def test_MultioutputRandomForest_predictions():
    X, Y = make_regression(n_samples=150, n_features=40, n_targets=10)
    Y = (Y - Y.mean(axis=0)) / Y.std(axis=0)

    base = RandomForestRegressor(max_depth=3, random_state=123).fit(X, Y)
    model = MultioutputRandomForest(max_depth=3, random_state=123).fit(X, Y)

    assert np.allclose(base.predict(X), model.predict(X))

def test_feature_importances__shape():
    X, Y = make_regression(n_samples=150, n_features=40, n_targets=10)
    Y = (Y - Y.mean(axis=0)) / Y.std(axis=0)

    random = MultioutputRandomForest(max_depth=3).fit(X, Y)
    boosted = MultioutputBoostedForest(max_depth=3).fit(X, Y)

    assert random.feature_importances_.shape == (40, 10)
    assert boosted.feature_importances_.shape == (40, 10)

def test_feature_importances__sum():
    X, Y = make_regression(n_samples=150, n_features=40, n_targets=10)
    Y = (Y - Y.mean(axis=0)) / Y.std(axis=0)

    random = MultioutputRandomForest(max_depth=3).fit(X, Y)
    boosted = MultioutputBoostedForest(max_depth=3).fit(X, Y)

    assert np.allclose(np.ones(10), random.feature_importances_.sum(axis=0))
    assert np.allclose(np.ones(10), boosted.feature_importances_.sum(axis=0))


from __future__ import annotations

from typing import Optional

import numpy as np

from sklearn.tree import DecisionTreeRegressor
from sklearn.utils.validation import _check_sample_weight, check_X_y

class MultioutputDecisionTree(DecisionTreeRegressor):
    """Tree regressor with per-target feature importances."""

    feature_importances_ = np.array([])

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        sample_weight: Optional[np.ndarray]=None,
        **_
    ) -> MultioutputDecisionTree:
        """
        Fit the regressor and calculate feature importances and absolute means.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The training inputs.
        Y : np.ndarray of shape (n_samples,) or (n_samples, n_outputs)
            The training targets.
        sample_weight : np.ndarray of shape (n_samples,) or None
            The sample weights. If None, all samples are weighted equally.

        Returns
        -------
        self : MultioutputTreeRegressor
            The fitted regressor.

        """

        X, y = check_X_y(X, y, multi_output=True, y_numeric=True)
        tree = super().fit(X, y, sample_weight, check_input=False).tree_
        
        paths = self.decision_path(X).toarray().astype(bool)
        n_samples = paths.sum(axis=0)
        n_nodes = len(n_samples)

        impurities = np.zeros((n_nodes, self.n_outputs_))
        for node in range(n_nodes):
            impurities[node] = y[paths[:, node]].var(axis=0)

        importances = np.zeros((X.shape[1], self.n_outputs_))
        for node in range(n_nodes):
            left = tree.children_left[node]
            right = tree.children_right[node]

            if left == right:
                continue

            gain = n_samples[node] * impurities[node]
            gain -= n_samples[left] * impurities[left]
            gain -= n_samples[right] * impurities[right]

            feature = tree.feature[node]
            importances[feature] += gain

        leaves = tree.children_left == tree.children_right
        outputs = tree.value[leaves, :, 0].T
        n_samples = n_samples[leaves]

        sums = importances.sum(axis=0)
        importances[:, sums != 0] /= sums[sums != 0]

        if self.n_outputs_ == 1:
            importances = importances.flatten()

        self.abs_means_: np.ndarray
        self.feature_importances_ = importances
        self.abs_means_ = (n_samples * outputs).sum(axis=1) / n_samples.sum()

        return self
